==> vmmagent-0.6.4-sources.jar: org/ow2/jasmine/vmm/agent/driver/libvirt/ImageCatalog.java <==
/**
 * JASMINe VMMapi: JASMINe Virtual Machine Management API
 * Copyright (C) 2009 France Telecom R&D
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: ImageCatalog.java 6173 2010-03-23 17:03:34Z dangtran $
 * --------------------------------------------------------------------------
 */
package org.ow2.jasmine.vmm.agent.driver.libvirt;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.management.ObjectName;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathFactory;

import org.apache.log4j.Logger;
import org.ow2.jasmine.vmm.agent.domain.ManagedResource;
import org.ow2.jasmine.vmm.agent.driver.util.RemoteExec;
import org.ow2.jasmine.vmm.agent.driver.util.RemoteExec.SshException;
import org.ow2.jasmine.vmm.agent.driver.virtimage.FileUtils;
import org.ow2.jasmine.vmm.agent.jmx.MBeanObjectNamer;
import org.ow2.jasmine.vmm.agent.main.AgentCommon;
import org.ow2.jasmine.vmm.api.BadVMPowerStateException;
import org.ow2.jasmine.vmm.api.IllegalOperationException;
import org.ow2.jasmine.vmm.api.InsufficientResourcesException;
import org.ow2.jasmine.vmm.api.VMConfigSpec;
import org.ow2.jasmine.vmm.api.VMMException;
import org.ow2.jasmine.vmm.api.VirtualMachineImageMXBean;
import org.ow2.jasmine.vmm.api.VirtualMachineImageStoreMXBean;
import org.ow2.jasmine.vmm.api.VnicIPSettings;
import org.ow2.jasmine.vmm.api.VirtualMachineMXBean.PowerState;
import org.ow2.jasmine.vmm.api.VnicIPSettings.IpAssignmentMode;
import org.w3c.dom.Document;

public class ImageCatalog extends ManagedResource implements VirtualMachineImageStoreMXBean {
    static Logger logger = Logger.getLogger(ImageCatalog.class);

    private String rootDirectory;

    private String storeServer;

    private RemoteExec.SshAuthInfo authInfo;

    private CopyOnWriteArrayList<VirtualMachineImage> vmList = new CopyOnWriteArrayList<VirtualMachineImage>();

    public ImageCatalog(final ObjectName objectName, final String rootDirectory, final String storeServer,
        final RemoteExec.SshAuthInfo authInfo) throws VMMException {
        super(objectName);
        this.storeServer = storeServer;
        this.rootDirectory = rootDirectory;
        if (rootDirectory.endsWith("/")) {
            this.rootDirectory = rootDirectory.substring(0, rootDirectory.length() - 1);
        }
        this.authInfo = authInfo;
        this.sync();
    }

    public String getRootDirectory() {
        return this.rootDirectory;
    }

    public void removeVMImageTemplate(final VirtualMachineImageMXBean vmImage) throws VMMException {
        String vmImageID = vmImage.getUUID();
        try {
            String command = "rm -fr " + this.rootDirectory + "/" + vmImageID + ".template";
            ImageCatalog.logger.debug("Removing VM image: " + command);
            RemoteExec.Result result = RemoteExec.commandAsRoot(this.storeServer, this.authInfo, command);
            if (result.exitCode != 0) {
                ImageCatalog.logger.error(result.error);
                throw new VMMException(result.error);
            }
        } catch (Exception ex) {
            throw new VMMException(ex.getMessage());
        }
    }

    synchronized void sync() throws VMMException {
        RemoteExec.Result result;
        try {
            String command = "ls -d1 " + this.rootDirectory + "/*.template";
            ImageCatalog.logger.debug("Listing VM images: " + command);
            result = RemoteExec.commandAsRoot(this.storeServer, this.authInfo, command);
        } catch (SshException ex) {
            throw new VMMException(ex.getMessage());
        }
        if (result.exitCode != 0) {
            ImageCatalog.logger.error(result.error);
            throw new VMMException(result.error);
        }
        String imageFileNames[] = result.output.split("\\s+");
        // add and register new images
        for (String fileName : imageFileNames) {
            String uuid = fileName.substring(0, fileName.length() - ".template".length());
            uuid = uuid.substring(1 + uuid.lastIndexOf("/"));
            if (this.lookUpByUUID(uuid) == null) {
                ObjectName objName = MBeanObjectNamer.makeVMImageName(uuid);
                try {
                    VirtualMachineImage image = new VirtualMachineImage(objName, uuid, fileName);
                    AgentCommon.getMBeanServer().registerMBean(image, image.getObjectName());
                    this.vmList.add(image);
                    ImageCatalog.logger.info("Added VMImage " + image.name + " (" + image.description + ")");
                } catch (Exception ex) {
                    ImageCatalog.logger.error("Failed to add VMImage MBean", ex);
                    continue;
                }

            }
        }
        // get rid of deleted images
        for (VirtualMachineImage image : this.vmList) {
            boolean match = false;
            for (String fileName : imageFileNames) {
                String uuid = fileName.substring(0, fileName.length() - ".template".length());
                uuid = uuid.substring(1 + uuid.lastIndexOf("/"));
                if (image.getUUID().equals(uuid)) {
                    match = true;
                    break;
                }
            }
            if (!match) {
                try {
                    AgentCommon.getMBeanServer().unregisterMBean(MBeanObjectNamer.makeVMImageName(image.getUUID()));
                } catch (Exception ex) {
                    ImageCatalog.logger.error("Failed to unregister VMImage MBean", ex);
                }
                this.vmList.remove(image);
            }
        }
    }

    public VirtualMachineImageMXBean newVMImageTemplate(final LibvirtVirtualMachine vm, final String name,
        final String description, final Map<String, String> metadata) throws InsufficientResourcesException,
        IllegalOperationException, BadVMPowerStateException, VMMException {

        if (vm.getHostMBean().getVMImageStore() != this) {
            throw new IllegalOperationException("VM is not attached to this ImageStore");
        }
        if (vm.getState() == PowerState.RUNNING) {
            throw new BadVMPowerStateException("Cannot make a template from a running VM");
        }
        String uuid = name + "-" + UUID.randomUUID();
        List<String> disks = vm.getDiskFiles();
        if (disks == null || disks.size() == 0) {
            throw new IllegalOperationException("No disks found on this VM");
        }
        RemoteExec.Result result = null;
        try {
            String command = this.rootDirectory + "/makeNewVMTemplate --vm " + vm.getNameLabel() + " --uuid " + uuid
                + " --name \"" + name + "\" --desc \"" + description + " \" " + " --disk " + disks.get(0);
            ImageCatalog.logger.debug("Launching command " + command);
            result = RemoteExec.commandAsRoot(this.storeServer, this.authInfo, command);
        } catch (Exception ex) {
            ImageCatalog.logger.error("SSH failure with host " + this.storeServer, ex);
            throw new VMMException(ex.toString());
        }
        if (result.exitCode != 0) {
            ImageCatalog.logger.error("makeNewVMTemplate failed: " + result.error);
            throw new VMMException(result.error);
        }

        this.sync();
        VirtualMachineImageMXBean image = this.lookUpByUUID(uuid);
        if (image == null) {
            throw new VMMException("Failed to locate image with uuid " + uuid);
        }
        return image;
    }

    public VirtualMachineImageMXBean lookUpByName(final String name) {
        for (VirtualMachineImage image : this.vmList) {
            if (image.getName().equals(name)) {
                return image;
            }
        }
        return null;
    }

    public VirtualMachineImageMXBean lookUpByUUID(final String uuid) {
        for (VirtualMachineImage image : this.vmList) {
            if (image.getUUID().equals(uuid)) {
                return image;
            }
        }
        return null;
    }

    public List<VirtualMachineImageMXBean> listVMImageTemplates() {
        ArrayList<VirtualMachineImageMXBean> result = new ArrayList<VirtualMachineImageMXBean>(this.vmList);
        return result;
    }

    public long getCapacityMB() {
        return FileUtils.getDiskCapacityMB(this.storeServer, this.authInfo, this.rootDirectory);
    }

    public long getFreeSpaceMB() {
        return FileUtils.getDiskFreeSpaceMB(this.storeServer, this.authInfo, this.rootDirectory);
    }

    public String getName() {
        return "ImageStore(" + this.rootDirectory + ")";
    }

    public class VirtualMachineImage extends ManagedResource implements VirtualMachineImageMXBean {
        private String name, uuid, description, dir;

        private VirtualMachineImage(final ObjectName objectName, final String uuid, final String dir) {
            super(objectName);
            this.uuid = uuid;
            this.dir = dir;
            try {
                this.readMetaData();
            } catch (Exception ex) {
                ImageCatalog.logger.error("Failed to read metatdata for VM image " + uuid, ex);
            }
        }

        private void readMetaData() throws Exception {
            RemoteExec.Result result;
            String command = "cat " + this.dir + "/metadata.xml";
            result = RemoteExec.commandAsRoot(ImageCatalog.this.storeServer, ImageCatalog.this.authInfo, command);
            if (result.exitCode != 0) {
                throw new Exception(result.error);
            }
            String metaData = result.output;
            Document xmlDocument;
            XPath xPath;
            xmlDocument = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(
                new ByteArrayInputStream(metaData.getBytes("UTF-8")));
            xPath = XPathFactory.newInstance().newXPath();
            XPathExpression xPathExpression = xPath.compile("/virtualMachineImage/description");
            this.description = (String) xPathExpression.evaluate(xmlDocument, XPathConstants.STRING);
            xPathExpression = xPath.compile("/virtualMachineImage/@name");
            this.name = (String) xPathExpression.evaluate(xmlDocument, XPathConstants.STRING);
        }

        public Map<String, String> getMetaData() {
            return null;
        }

        public String getName() {
            return this.name;
        }

        public String getDescription() {
            return this.description;
        }

        public String getUUID() {
            return this.uuid;
        }

        public void newVM(final VMConfigSpec vmSpec, final String vmName, final LibvirtHost host) throws VMMException {
            RemoteExec.Result result;
            String command = this.dir + "/createVM --name " + vmName + " --ncpu " + vmSpec.getNumVCPU() + " --memoryMB "
                + vmSpec.getMemorySizeMB() + " --diskSizeMB " + vmSpec.getDiskSizeMB() + " --targetDir "
                + host.serverPool.getSharedDiskStoreDir();
            if (vmSpec.getGuestOsHostName() != null && !vmSpec.getGuestOsHostName().equals("")) {
                command += " --custom --hostname " + vmSpec.getGuestOsHostName();
            }
            // XXX configure only first NIC
            if (vmSpec.getVnicIpSettingsList() != null) {
                VnicIPSettings nic0Settings = vmSpec.getVnicIpSettingsList().get(0);
                if (nic0Settings.getIpAssignmentMode() == IpAssignmentMode.DHCP) {
                    command += " --custom --net eth0/dhcp";
                } else if (nic0Settings.getIpAssignmentMode() == IpAssignmentMode.FIXED) {
                    command += " --custom --net eth0/static/" + nic0Settings.getIpAddress() + "/"
                        + nic0Settings.getSubnetMask() + "/" + nic0Settings.getGateway();
                }
            }

            ImageCatalog.logger.debug("VMImage " + this.name + " NewVM: launching command " + command);

            try {
                result = RemoteExec.commandAsRoot(host.getHostName(), host.getSshAuthInfo(), command);
            } catch (RemoteExec.SshException ex) {
                throw new VMMException("VM creation failed: SSH connection failure", ex);
            }
            if (result.exitCode != 0) {
                throw new VMMException("VM creation failed: " + result.error);
            }
        }
    }
}

==> vmmagent-0.6.5-sources.jar: org/ow2/jasmine/vmm/agent/driver/libvirt/ImageCatalog.java <==
/**
 * JASMINe VMMapi: JASMINe Virtual Machine Management API
 * Copyright (C) 2009 France Telecom R&D
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: ImageCatalog.java 7882 2011-03-30 16:46:31Z dangtran $
 * --------------------------------------------------------------------------
 */
package org.ow2.jasmine.vmm.agent.driver.libvirt;

import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.CopyOnWriteArrayList;

import javax.management.ObjectName;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathFactory;

import org.apache.log4j.Logger;
import org.ow2.jasmine.vmm.agent.domain.ManagedResource;
import org.ow2.jasmine.vmm.agent.driver.util.RemoteExec;
import org.ow2.jasmine.vmm.agent.driver.util.RemoteExec.SshException;
import org.ow2.jasmine.vmm.agent.driver.virtimage.FileUtils;
import org.ow2.jasmine.vmm.agent.jmx.MBeanObjectNamer;
import org.ow2.jasmine.vmm.agent.main.AgentCommon;
import org.ow2.jasmine.vmm.api.BadVMPowerStateException;
import org.ow2.jasmine.vmm.api.IllegalOperationException;
import org.ow2.jasmine.vmm.api.InsufficientResourcesException;
import org.ow2.jasmine.vmm.api.VMConfigSpec;
import org.ow2.jasmine.vmm.api.VMMException;
import org.ow2.jasmine.vmm.api.VirtualMachineImageMXBean;
import org.ow2.jasmine.vmm.api.VirtualMachineImageStoreMXBean;
import org.ow2.jasmine.vmm.api.VnicIPSettings;
import org.ow2.jasmine.vmm.api.VnicIPSettings.IpAssignmentMode;
import org.w3c.dom.Document;

public class ImageCatalog extends ManagedResource implements VirtualMachineImageStoreMXBean {
    static Logger logger = Logger.getLogger(ImageCatalog.class);

    private String rootDirectory;

    private String storeServer;

    private RemoteExec.SshAuthInfo authInfo;

    private CopyOnWriteArrayList<VirtualMachineImage> vmList = new CopyOnWriteArrayList<VirtualMachineImage>();

    public ImageCatalog(final ObjectName objectName, final String rootDirectory, final String storeServer,
        final RemoteExec.SshAuthInfo authInfo) throws VMMException {
        super(objectName);
        this.storeServer = storeServer;
        this.rootDirectory = rootDirectory;
        if (rootDirectory.endsWith("/")) {
            this.rootDirectory = rootDirectory.substring(0, rootDirectory.length() - 1);
        }
        this.authInfo = authInfo;
        this.sync();
    }

    public String getRootDirectory() {
        return this.rootDirectory;
    }

    public void removeVMImageTemplate(final VirtualMachineImageMXBean vmImage) throws VMMException {
        String vmImageID = vmImage.getUUID();
        try {
            String command = "rm -fr " + this.rootDirectory + "/" + vmImageID + ".template";
            ImageCatalog.logger.debug("Removing VM image: " + command);
            RemoteExec.Result result = RemoteExec.commandAsRoot(this.storeServer, this.authInfo, command);
            if (result.exitCode != 0) {
                ImageCatalog.logger.error(result.error);
                throw new VMMException(result.error);
            }
        } catch (Exception ex) {
            throw new VMMException(ex.getMessage());
        }
    }

    synchronized void sync() throws VMMException {
        RemoteExec.Result result;
        try {
            String command = "ls -d1 " + this.rootDirectory + "/*.template";
            ImageCatalog.logger.debug("Listing VM images: " + command);
            result = RemoteExec.commandAsRoot(this.storeServer, this.authInfo, command);
        } catch (SshException ex) {
            throw new VMMException(ex.getMessage());
        }
        if (result.exitCode != 0) {
            ImageCatalog.logger.error(result.error);
            throw new VMMException(result.error);
        }
        String imageFileNames[] = result.output.split("\\s+");
        // add and register new images
        for (String fileName : imageFileNames) {
            String uuid = fileName.substring(0, fileName.length() - ".template".length());
            uuid = uuid.substring(1 + uuid.lastIndexOf("/"));
            if (this.lookUpByUUID(uuid) == null) {
                ObjectName objName = MBeanObjectNamer.makeVMImageName(uuid);
                try {
                    VirtualMachineImage image = new VirtualMachineImage(objName, uuid, fileName);
                    AgentCommon.getMBeanServer().registerMBean(image, image.getObjectName());
                    this.vmList.add(image);
                    ImageCatalog.logger.info("Added VMImage " + image.name + " (" + image.description + ")");
                } catch (Exception ex) {
                    ImageCatalog.logger.error("Failed to add VMImage MBean", ex);
                    continue;
                }

            }
        }
        // get rid of deleted images
        for (VirtualMachineImage image : this.vmList) {
            boolean match = false;
            for (String fileName : imageFileNames) {
                String uuid = fileName.substring(0, fileName.length() - ".template".length());
                uuid = uuid.substring(1 + uuid.lastIndexOf("/"));
                if (image.getUUID().equals(uuid)) {
                    match = true;
                    break;
                }
            }
            if (!match) {
                try {
                    AgentCommon.getMBeanServer().unregisterMBean(MBeanObjectNamer.makeVMImageName(image.getUUID()));
                } catch (Exception ex) {
                    ImageCatalog.logger.error("Failed to unregister VMImage MBean", ex);
                }
                this.vmList.remove(image);
            }
        }
    }

    public VirtualMachineImageMXBean newVMImageTemplate(final LibvirtVirtualMachine vm, final String name,
        final String description, final Map<String, String> metadata) throws InsufficientResourcesException,
        IllegalOperationException, BadVMPowerStateException, VMMException {

        if (vm.getHostMBean().getVMImageStore() != this) {
            throw new IllegalOperationException("VM is not attached to this ImageStore");
        }
        String uuid = "vmi-" + UUID.randomUUID();
        List<String> disks = vm.getDiskFiles();
        if (disks == null || disks.size() == 0) {
            throw new IllegalOperationException("No disks found on this VM");
        }
        RemoteExec.Result result = null;
        try {
            String command = this.rootDirectory + "/makeNewVMTemplate --vm " + vm.getNameLabel() + " --uuid " + uuid
                + " --name \"" + name + "\" --desc \"" + description + " \" " + " --disk " + disks.get(0);
            ImageCatalog.logger.debug("Launching command " + command);
            result = RemoteExec.commandAsRoot(this.storeServer, this.authInfo, command);
        } catch (Exception ex) {
            ImageCatalog.logger.error("SSH failure with host " + this.storeServer, ex);
            throw new VMMException(ex.toString());
        }
        if (result.exitCode != 0) {
            ImageCatalog.logger.error("makeNewVMTemplate failed: " + result.error);
            throw new VMMException(result.error);
        }

        this.sync();
        VirtualMachineImageMXBean image = this.lookUpByUUID(uuid);
        if (image == null) {
            throw new VMMException("Failed to locate image with uuid " + uuid);
        }
        return image;
    }

    public VirtualMachineImageMXBean lookUpByName(final String name) {
        for (VirtualMachineImage image : this.vmList) {
            if (image.getName().equals(name)) {
                return image;
            }
        }
        return null;
    }

    public VirtualMachineImageMXBean lookUpByUUID(final String uuid) {
        for (VirtualMachineImage image : this.vmList) {
            if (image.getUUID().equals(uuid)) {
                return image;
            }
        }
        return null;
    }

    public List<VirtualMachineImageMXBean> listVMImageTemplates() {
        ArrayList<VirtualMachineImageMXBean> result = new ArrayList<VirtualMachineImageMXBean>(this.vmList);
        return result;
    }

    public long getCapacityMB() {
        return FileUtils.getDiskCapacityMB(this.storeServer, this.authInfo, this.rootDirectory);
    }

    public long getFreeSpaceMB() {
        return FileUtils.getDiskFreeSpaceMB(this.storeServer, this.authInfo, this.rootDirectory);
    }

    public String getName() {
        return "ImageStore(" + this.rootDirectory + ")";
    }

    public class VirtualMachineImage extends ManagedResource implements VirtualMachineImageMXBean {
        private String name, uuid, description, dir;

        private VirtualMachineImage(final ObjectName objectName, final String uuid, final String dir) {
            super(objectName);
            this.uuid = uuid;
            this.dir = dir;
            try {
                this.readMetaData();
            } catch (Exception ex) {
                ImageCatalog.logger.error("Failed to read metatdata for VM image " + uuid, ex);
            }
        }

        private void readMetaData() throws Exception {
            RemoteExec.Result result;
            String command = "cat " + this.dir + "/metadata.xml";
            result = RemoteExec.commandAsRoot(ImageCatalog.this.storeServer, ImageCatalog.this.authInfo, command);
            if (result.exitCode != 0) {
                throw new Exception(result.error);
            }
            String metaData = result.output;
            Document xmlDocument;
            XPath xPath;
            xmlDocument = DocumentBuilderFactory.newInstance().newDocumentBuilder()
                .parse(new ByteArrayInputStream(metaData.getBytes("UTF-8")));
            xPath = XPathFactory.newInstance().newXPath();
            XPathExpression xPathExpression = xPath.compile("/virtualMachineImage/description");
            this.description = (String) xPathExpression.evaluate(xmlDocument, XPathConstants.STRING);
            xPathExpression = xPath.compile("/virtualMachineImage/@name");
            this.name = (String) xPathExpression.evaluate(xmlDocument, XPathConstants.STRING);
        }

        public Map<String, String> getMetaData() {
            return null;
        }

        public String getName() {
            return this.name;
        }

        public String getDescription() {
            return this.description;
        }

        public String getUUID() {
            return this.uuid;
        }

        public void newVM(final VMConfigSpec vmSpec, final String vmName, final LibvirtHost host) throws VMMException {
            RemoteExec.Result result;
            String command = this.dir + "/createVM --name " + vmName + " --ncpu " + vmSpec.getNumVCPU() + " --memoryMB "
                + vmSpec.getMemorySizeMB() + " --diskSizeMB " + vmSpec.getDiskSizeMB() + " --targetDir "
                + host.serverPool.getSharedDiskStoreDir();
            if (vmSpec.getGuestOsHostName() != null && !vmSpec.getGuestOsHostName().equals("")) {
                command += " --hostname " + vmSpec.getGuestOsHostName();
            }
            // XXX configure only first NIC
            if (vmSpec.getVnicIpSettingsList() != null && vmSpec.getVnicIpSettingsList().size() > 0) {
                VnicIPSettings nic0Settings = vmSpec.getVnicIpSettingsList().get(0);
                if (nic0Settings.getIpAssignmentMode() == IpAssignmentMode.DHCP) {
                    command += "  --net eth0/dhcp";
                } else if (nic0Settings.getIpAssignmentMode() == IpAssignmentMode.FIXED) {
                    command += "  --net eth0/static/" + nic0Settings.getIpAddress() + "/" + nic0Settings.getSubnetMask() + "/"
                        + nic0Settings.getGateway();
                }
            }
            if (vmSpec.getGuestOsCustomizationParams() != null) {
                for (String param : vmSpec.getGuestOsCustomizationParams().keySet()) {
                    if (param.equals("sshkey")) {
                        command += " --sshkey \"" + vmSpec.getGuestOsCustomizationParams().get(param) + "\"";
                    } else {
                        command += " --userdata \"" + param + "\" \"" + vmSpec.getGuestOsCustomizationParams().get(param)
                            + "\"";
                    }
                }
            }

            ImageCatalog.logger.debug("VMImage " + this.name + " NewVM: launching command " + command);

            try {
                result = RemoteExec.commandAsRoot(host.getHostName(), host.getSshAuthInfo(), command);
            } catch (RemoteExec.SshException ex) {
                throw new VMMException("VM creation failed: SSH connection failure", ex);
            }
            if (result.exitCode != 0) {
                throw new VMMException("VM creation failed: " + result.error);
            }
        }
    }
}
